/*-
 * #%L
 * ImgLib2: a general-purpose, multidimensional image processing library.
 * %%
 * Copyright (C) 2009 - 2024 Tobias Pietzsch, Stephan Preibisch, Stephan Saalfeld,
 * John Bogovic, Albert Cardona, Barry DeZonia, Christian Dietz, Jan Funke,
 * Aivar Grislis, Jonathan Hale, Grant Harris, Stefan Helfrich, Mark Hiner,
 * Martin Horn, Steffen Jaensch, Lee Kamentsky, Larry Lindsey, Melissa Linkert,
 * Mark Longair, Brian Northan, Nick Perry, Curtis Rueden, Johannes Schindelin,
 * Jean-Yves Tinevez and Michael Zinsmaier.
 * %%
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * #L%
 */
package net.imglib2.type.logic;

import static org.junit.Assert.assertEquals;

import java.util.Random;

import net.imglib2.img.array.ArrayImg;
import net.imglib2.img.array.ArrayImgFactory;
import net.imglib2.img.array.ArrayImgs;
import net.imglib2.img.basictypeaccess.array.ArrayDataAccess;
import net.imglib2.img.basictypeaccess.array.LongArray;

import org.junit.Test;

/**
 * Tests that we do not allocate overly large storage arrays for BitType images.
 *
 * @author Gabriel Einsdorf
 */
public class BitTypeArrayImgSizeTest
{

	@Test
	public void testOverSizedBitImageExplicitStorageArray()
	{
		final int numLongs = ( int ) Math.ceil( 100 * 100 / 64d );
		final ArrayImg< BitType, LongArray > img = ArrayImgs.bits( new LongArray( new long[ numLongs ] ), 100, 100 );
		final Random r = new Random( 42l );
		img.forEach( p -> {
			p.set( r.nextBoolean() );
		} );

		final long[] storage = ( long[] ) ( ( ArrayDataAccess< ? > ) img.update( null ) ).getCurrentStorageArray();
		final long[] sizes = new long[ img.numDimensions() ];
		img.dimensions( sizes );

		assertEquals( storage.length, numLongs );
	}

	@Test
	public void testOverSizedBitImageImgFactory()
	{
		final ArrayImgFactory< BitType > factory = new ArrayImgFactory<>( new BitType() );

		// evenly dividable by 64
		final int numLongs = 1000 * 1000 / 64;
		final ArrayImg< BitType, ? > img = factory.create( 1000, 1000 );

		final Random r = new Random( 42l );
		img.forEach( p -> {
			p.set( r.nextBoolean() );
		} );

		final long[] storage = ( long[] ) ( ( ArrayDataAccess< ? > ) img.update( null ) ).getCurrentStorageArray();

		final long[] sizes = new long[ img.numDimensions() ];
		img.dimensions( sizes );
		assertEquals( numLongs, storage.length );
	}

	@Test
	public void testOverSizedBitImageArrayImgs()
	{
		final int numLongs = ( int ) Math.ceil( 100 * 100 / 64d );
		final ArrayImg< BitType, LongArray > img = ArrayImgs.bits( 100, 100 );

		final Random r = new Random( 42l );
		img.forEach( p -> {
			p.set( r.nextBoolean() );
		} );

		final long[] storage = ( long[] ) ( ( ArrayDataAccess< ? > ) img.update( null ) ).getCurrentStorageArray();
		final long[] sizes = new long[ img.numDimensions() ];
		img.dimensions( sizes );

		assertEquals( numLongs, storage.length );
	}
}
